/* WZeta assembler & simulator
   WZeta Project https://wzeta.idletime.tokyo/
   Naoki Hirayama
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "wzsim.h"

#define LINE_SIZE (2048)
#define NEM_SIZE (32)
#define WORD_SIZE (32)
#define MAX_LBL (1024)
#define MAX_DEF (256)

unsigned int *PCODE;
unsigned long long *PDEBUG;
char **LDEBUG;
int adr_line;

static char lbl[MAX_LBL][NEM_SIZE];
static unsigned short lbl_adr[MAX_LBL];
static int lbl_ptr;

static char def[MAX_DEF][WORD_SIZE];
static unsigned short def_num[MAX_DEF];
static int def_ptr;

static char buf[LINE_SIZE];
static unsigned long long debug;
static int err_line;

typedef struct {
    char w0[WORD_SIZE];
    char w1[WORD_SIZE];
    char w2[WORD_SIZE];
    char w3[WORD_SIZE];
    unsigned short s0, s1, s2, s3;
} asmwords;

void exit_msg(char* msg) {
    fprintf(fp,"%s\n",msg);
    exit(-1);
}

void exit_msg_s(char* msg,char *s) {
    fprintf(fp,"%s %s\n",msg,s);
    exit(-1);
}

#define EXIT_SYNTAX_ERROR() exit_syntax_error(__LINE__)

void exit_syntax_error(int n) {
    fprintf(fp,"ERROR: syntax error(%d) in %d\n%s\n",n,err_line,buf);
    exit(-1);
}

void exit_range_over() {
    fprintf(fp,"ERROR: range over in %d\n%s\n",err_line,buf);
    exit(-1);
}

void delete_space(char *p) {
    char *q = p;
    do {
        if(*p!=' ' && *p!='\t') *q++ = *p;
    } while(*p++!='\0');
}

unsigned short get_address(char* w) {
    int i,count;
    unsigned short jmp = 0xffff;
	for(i=0,count=0 ; i<lbl_ptr ; i++) {
        if(strcmp(w,lbl[i])==0) {
              jmp = lbl_adr[i];
			  count++;
		}
	}
	if(count > 1) exit_msg("label conflict.");
	if(count == 0) exit_msg_s("label not found.",w);
	if(jmp > 0x3fff) exit_msg("label out of range.");
    return jmp;
}

unsigned short asm_oprx(char *w,int load) {
    unsigned long num;
    unsigned short r = 0;
    char *endptr;
    if(*w=='R') {
        r = 0x8000;
        w++;
    }
    if(*w=='$') {
        int i;
        int count = 0;
        num = 0xffffffff;
        for(i=0 ; i<def_ptr ; i++) {
            if(strncmp(w,def[i],WORD_SIZE)==0) {
                num = def_num[i];
                count++;
            }
        }
        if(count>1) exit_msg_s("double define.",w);
    } else {
        errno = 0;
        endptr = w;
       if(*w=='\0') EXIT_SYNTAX_ERROR();
        num = strtoul(w,&endptr,0);
       if(w==endptr) EXIT_SYNTAX_ERROR();
       if(*endptr!='\0') EXIT_SYNTAX_ERROR();
       if(errno) EXIT_SYNTAX_ERROR();
   }
   if(load) {
       if(r==0 && num>0xff) exit_range_over();
       if(r!=0 && num>0x7f) exit_range_over();
   } else {
       if(num > 0x7f) exit_range_over();
   }
   r |= num;
   return r;
}

unsigned short asm_opr(char *w) {
    return asm_oprx(w,0);
}

unsigned short asm_opr_load(char *w) {
    return asm_oprx(w,1);
}

unsigned short opr_7bit(unsigned short n) {
    if(n & 0x7f00) EXIT_SYNTAX_ERROR();
    if(n&0x8000) n = 0x80 | (0x7F&n);
    return n;
}

int is_lbl(char *p,unsigned long adr, int pass) {
    long i,num;
    char *endptr;

    for(i=0 ; i<LINE_SIZE && p[i]!=':' && p[i]!='\0' ; i++)
 	    lbl[lbl_ptr][i] = p[i];
    if(p[i]!=':') return 1;

    if(pass==1) {
        lbl_adr[lbl_ptr] = adr;
     	lbl[lbl_ptr++][i] = '\0';
        if(lbl_ptr>=MAX_LBL) exit_msg("too many lbls");
    }

    p = p+i+1;
    while(*p=='\0' || *p=='\n' || *p=='\r' || *p==' ' || *p=='\t') {
        if(*p=='\0' || *p=='\n' || *p=='\r') return 0;
        p++;
    }
    errno = 0;
    endptr = p;
    num = strtol(p,&endptr,0);
   if(p==endptr || errno) {
       fprintf(fp,"ERROR: label address in %d\n",err_line);
       exit(-1);
   }
   if(*endptr!='\0' && *endptr!='\n' && *endptr!='\r' && *endptr!=' ' && *endptr!='\t') {
       fprintf(fp,"ERROR: label address in %d\n",err_line);
       exit(-1);
   }
   num = num - adr_line;
   if(num<0 || num>0x3fff) {
        fprintf(fp,"ERROR: label address in %d\n",err_line);
        exit(-1);
    }
    return num;
}

int deb_line(char* line,int r) {
    char *p = line;
    char *q;
    debug = 0;
    delete_space(p);
    char w[LINE_SIZE];

    do {
        q = w;
        while(*p!='\0' && *p!='\n' && *p!='\r' && *p!=';') *q++ = *p++;
        *q='\0';
        if(w[0]=='@' && w[1]=='R') debug |= 1 << atoi(w+2);
    } while(*p++ != '\0');
    PDEBUG[adr_line] = debug;
    return r;
}

int asm_line(asmwords* aw,char *line) {
    char *p = line;
    char *q;

    debug = 0;

    aw->w0[0] = aw->w1[0] = aw->w2[0] = aw->w3[0] = '\0';

    while(*p==' ' || *p=='\t') if(*p++=='\0') EXIT_SYNTAX_ERROR();
    q = aw->w0;
    while( (*p>='A' && *p<='Z')  || *p=='&' || *p=='0' || *p=='1' || *p=='2' ) *q++ = *p++;
    *q = '\0';
    while(*p==' ' || *p=='\t') p++;
    if(*p=='\0' || *p=='\n' || *p=='\r') return 1;
    if(*p==';') return deb_line(p,1);

    q = aw->w1;
    while( *p!=',' && *p!='\0' && *p!=';') {
        if(*p!=' ' && *p!='\t' && *p!='\n' && *p!='\r') *q++ = *p;
        p++;
    }
    *q = '\0';
    if(*p=='\0' || *p=='\n' || *p=='\r') return 2;
    if(*p++==';') return deb_line(p,2);

    q = aw->w2;
    while( *p!=',' && *p!='\0' && *p!=';') {
        if(*p!=' ' && *p!='\t' && *p!='\n' && *p!='\r') *q++ = *p;
        p++;
    }
    *q = '\0';
    if(*p=='\0' || *p=='\n' || *p=='\r') return 3;
    if(*p++==';') return deb_line(p,3);

    q = aw->w3;
    while( *p!=',' && *p!='\0' && *p!=';') {
        if(*p!=' ' && *p!='\t' && *p!='\n' && *p!='\r') *q++ = *p;
        p++;
    }
    *q = '\0';
    if(*p=='\0' || *p=='\n' || *p=='\r') return 4;
    if(*p++==';') return deb_line(p,4);

    return 4;
}

void wz_asm(const char* asmfile) {
    int i,c,n;
    FILE *fasm;
    unsigned int code;
    asmwords aw;
    unsigned int x,y,z;
        
    /* PASS1 */
    if((fasm=fopen(asmfile,"r"))==NULL) exit_msg("Can't open file");
	adr_line = lbl_ptr = def_ptr = err_line = 0;

    while(fgets(buf,LINE_SIZE,fasm)) {
		err_line++;
        if(buf[0]=='#') continue;
        if(buf[0]=='$') {
            char *p = buf;
            char *q = def[def_ptr];
            char *endptr;
            unsigned long num;

            while(*p!=' ' && *p!='\t' && *p!='\r' && *p!='\n' && *p!='\0') *q++ = *p++;
            if(*p=='\r' || *p=='\n' || *p=='\0') continue;
            *q = '\0';
            while(*p==' ' || *p=='\t') p++;

            errno = 0;
            endptr = p;
            num = strtol(p,&endptr,0);
            if(p==endptr || errno) {
                fprintf(fp,"ERROR: invalid define in %d\n",err_line);
                fprintf(fp,buf);
                exit(-1);
            }
            if(*endptr!='\0' && *endptr!='\n' && *endptr!='\r' && *endptr!=' ' && *endptr!='\t') {
                fprintf(fp,"ERROR: invalid define in %d\n",err_line);
                fprintf(fp,buf);
                exit(-1);
            }
            def_num[def_ptr++] = num;
            continue;
        }
		delete_space(buf);
		adr_line += is_lbl(buf,adr_line,1);
    }
    fclose(fasm);

    PCODE = (unsigned int*)calloc(adr_line,sizeof(unsigned int));
    PDEBUG = (unsigned long long*)calloc(adr_line,sizeof(unsigned long long));
    if(PCODE==NULL || PDEBUG==NULL) exit_msg("Out of memory");

    LDEBUG = (char**)calloc(adr_line,sizeof(char*));
    for(i=0 ; i<adr_line ; i++) {
        LDEBUG[i] = (char*)calloc(LINE_SIZE,sizeof(char));
        if(LDEBUG[i]==NULL) exit_msg("Out of memory");
    }

    /* PASS 2 */
    if((fasm=fopen(asmfile,"r"))==NULL) exit_msg("Can't open file");
    adr_line = err_line = 0;

    while(fgets(buf,LINE_SIZE,fasm)) {
		err_line++;
		if(buf[0]=='#' || buf[0]=='$') continue;
		c=0;
		for(i=0 ; buf[i] != '\0' ; i++) if(buf[i] == ':') c=1;
        if(c == 1) {
            if(mode&M_STDOUT) fprintf(fp,"%s",buf);
    		adr_line += is_lbl(buf,adr_line,2);
            continue;
        }
		code = 0;
        strcpy(LDEBUG[adr_line],buf);
        n = asm_line(&aw,buf);
   
//      printf("(%s) (%s) (%s) (%s)\n",aw.w0,aw.w1,aw.w2,aw.w3);

        if(strcmp(aw.w0,"NOP")==0) code = 0x80;

        if(strcmp(aw.w0,"LOAD")==0) {
            if( n != 3 ) EXIT_SYNTAX_ERROR();
            z = opr_7bit(asm_opr(aw.w1));
            if( (z&0x80)==0 ) EXIT_SYNTAX_ERROR();

            x = asm_opr_load(aw.w2);
            if(x & 0x8000) {
                x = 0x80 | (0x7F&x);
                code = (z<<24) | (0x00<<16) | x;
            } else {
                if(x>255) exit_msg("LOAD constant range over");
                y = x/2;
                c = x&1;
                x = y;
                code = (z<<24) | (0x00<<16) | (y<<8) | x | (c<<16);
            }
        }

        if(strcmp(aw.w0,"ADD")==0) {
            if( n != 4 ) EXIT_SYNTAX_ERROR();
            z = opr_7bit(asm_opr(aw.w1));
            if( (z&0x80)==0 ) EXIT_SYNTAX_ERROR();
            x = opr_7bit(asm_opr(aw.w2));
            y = opr_7bit(asm_opr(aw.w3));
            code = (z<<24) | (0x00<<16) | (y<<8) | x;
        }

        if(strcmp(aw.w0,"ADDC")==0) {
            if( n != 4 ) EXIT_SYNTAX_ERROR();
            z = opr_7bit(asm_opr(aw.w1));
            if( (z&0x80)==0 ) EXIT_SYNTAX_ERROR();
            x = opr_7bit(asm_opr(aw.w2));
            y = opr_7bit(asm_opr(aw.w3));
            code = (z<<24) | (0x04<<16) | (y<<8) | x;
        }

        if(strcmp(aw.w0,"SUB")==0) {
            if( n != 4 ) EXIT_SYNTAX_ERROR();
            z = opr_7bit(asm_opr(aw.w1));
            if( (z&0x80)==0 ) EXIT_SYNTAX_ERROR();
            x = opr_7bit(asm_opr(aw.w2));
            y = opr_7bit(asm_opr(aw.w3));
            code = (z<<24) | (0x03<<16) | (y<<8) | x;
        }

        if(strcmp(aw.w0,"SUBC")==0) {
            if( n != 4 ) EXIT_SYNTAX_ERROR();
            z = opr_7bit(asm_opr(aw.w1));
            if( (z&0x80)==0 ) EXIT_SYNTAX_ERROR();
            x = opr_7bit(asm_opr(aw.w2));
            y = opr_7bit(asm_opr(aw.w3));
            code = (z<<24) | (0x06<<16) | (y<<8) | x;
        }

        if(strcmp(aw.w0,"AND")==0) {
            if( n != 4 ) EXIT_SYNTAX_ERROR();
            z = opr_7bit(asm_opr(aw.w1));
            if( (z&0x80)==0 ) EXIT_SYNTAX_ERROR();
            x = opr_7bit(asm_opr(aw.w2));
            y = opr_7bit(asm_opr(aw.w3));
            code = (z<<24) | (0x41<<16) | (y<<8) | x;
        }

        if(strcmp(aw.w0,"OR")==0) {
            if( n != 4 ) EXIT_SYNTAX_ERROR();
            z = opr_7bit(asm_opr(aw.w1));
            if( (z&0x80)==0 ) EXIT_SYNTAX_ERROR();
            x = opr_7bit(asm_opr(aw.w2));
            y = opr_7bit(asm_opr(aw.w3));
            code = (z<<24) | (0x43<<16) | (y<<8) | x;
        }

        if(strcmp(aw.w0,"XOR")==0) {
            if( n != 4 ) EXIT_SYNTAX_ERROR();
            z = opr_7bit(asm_opr(aw.w1));
            if( (z&0x80)==0 ) EXIT_SYNTAX_ERROR();
            x = opr_7bit(asm_opr(aw.w2));
            y = opr_7bit(asm_opr(aw.w3));
            code = (z<<24) | (0x42<<16) | (y<<8) | x;
        }

        if(strcmp(aw.w0,"TEST")==0) {
            if( n != 3 ) EXIT_SYNTAX_ERROR();
            x = opr_7bit(asm_opr(aw.w1));
            y = opr_7bit(asm_opr(aw.w2));
            code = (0x41<<16) | (y<<8) | x;
        }

        if(strcmp(aw.w0,"COMPARE")==0) {
            if( n != 3 ) EXIT_SYNTAX_ERROR();
            x = opr_7bit(asm_opr(aw.w1));
            y = opr_7bit(asm_opr(aw.w2));
            code = (0x03<<16) | (y<<8) | x;
        }

        if(strcmp(aw.w0,"COMPAREC")==0) {
            if( n != 3 ) EXIT_SYNTAX_ERROR();
            x = opr_7bit(asm_opr(aw.w1));
            y = opr_7bit(asm_opr(aw.w2));
            code = (0x06<<16) | (y<<8) | x;
        }

        if(strcmp(aw.w0,"JUMP")==0) {
            if( n!=2 && n!=3 ) EXIT_SYNTAX_ERROR();
            if( n==2 ) {
                x = get_address(aw.w1);
                code = (0x10<<16) | ((x&0x3f80)<<1) | (x&0x7f);
            } else {
                y = opr_7bit(asm_opr(aw.w1));
                x = opr_7bit(asm_opr(aw.w2));
                if(y&0x80) EXIT_SYNTAX_ERROR();
                code = (0x10<<16) | y<<8 | x;
            }
        }

        if(strcmp(aw.w0,"JUMPC0")==0) {
            if( n!=2 && n!=3 ) EXIT_SYNTAX_ERROR();
            if( n==2 ) {
                x = get_address(aw.w1);
                code = (0x12<<16) | ((x&0x3f80)<<1) | (x&0x7f);
            } else {
                y = opr_7bit(asm_opr(aw.w1));
                x = opr_7bit(asm_opr(aw.w2));
                if(y&0x80) EXIT_SYNTAX_ERROR();
                code = (0x12<<16) | y<<8 | x;
            }
        }

        if(strcmp(aw.w0,"JUMPC1")==0) {
            if( n!=2 && n!=3 ) EXIT_SYNTAX_ERROR();
            if( n==2 ) {
                x = get_address(aw.w1);
                code = (0x13<<16) | ((x&0x3f80)<<1) | (x&0x7f);
            } else {
                y = opr_7bit(asm_opr(aw.w1));
                x = opr_7bit(asm_opr(aw.w2));
                if(y&0x80) EXIT_SYNTAX_ERROR();
                code = (0x13<<16) | y<<8 | x;
            }
        }

        if(strcmp(aw.w0,"JUMPNZ")==0) {
            if( n!=2 && n!=3 ) EXIT_SYNTAX_ERROR();
            if( n==2 ) {
                x = get_address(aw.w1);
                code = (0x14<<16) | ((x&0x3f80)<<1) | (x&0x7f);
            } else {
                y = opr_7bit(asm_opr(aw.w1));
                x = opr_7bit(asm_opr(aw.w2));
                if(y&0x80) EXIT_SYNTAX_ERROR();
                code = (0x14<<16) | y<<8 | x;
            }
        }

        if(strcmp(aw.w0,"JUMPZ")==0) {
            if( n!=2 && n!=3 ) EXIT_SYNTAX_ERROR();
            if( n==2 ) {
                x = get_address(aw.w1);
                code = (0x15<<16) | ((x&0x3f80)<<1) | (x&0x7f);
            } else {
                y = opr_7bit(asm_opr(aw.w1));
                x = opr_7bit(asm_opr(aw.w2));
                if(y&0x80) EXIT_SYNTAX_ERROR();
                code = (0x15<<16) | y<<8 | x;
            }
        }

        if(strcmp(aw.w0,"CALL")==0) {
            if( n!=2 && n!=3 ) EXIT_SYNTAX_ERROR();
            if( n==2 ) {
                x = get_address(aw.w1);
                code = (0x18<<16) | ((x&0x3f80)<<1) | (x&0x7f);
            } else {
                y = opr_7bit(asm_opr(aw.w1));
                x = opr_7bit(asm_opr(aw.w2));
                if(y&0x80) EXIT_SYNTAX_ERROR();
                code = (0x18<<16) | y<<8 | x;
            }
        }

        if(strcmp(aw.w0,"CALLC0")==0) {
            if( n!=2 && n!=3 ) EXIT_SYNTAX_ERROR();
            if( n==2 ) {
                x = get_address(aw.w1);
                code = (0x1A<<16) | ((x&0x3f80)<<1) | (x&0x7f);
            } else {
                y = opr_7bit(asm_opr(aw.w1));
                x = opr_7bit(asm_opr(aw.w2));
                if(y&0x80) EXIT_SYNTAX_ERROR();
                code = (0x1A<<16) | y<<8 | x;
            }
        }

        if(strcmp(aw.w0,"CALLC1")==0) {
            if( n!=2 && n!=3 ) EXIT_SYNTAX_ERROR();
            if( n==2 ) {
                x = get_address(aw.w1);
                code = (0x1B<<16) | ((x&0x3f80)<<1) | (x&0x7f);
            } else {
                y = opr_7bit(asm_opr(aw.w1));
                x = opr_7bit(asm_opr(aw.w2));
                if(y&0x80) EXIT_SYNTAX_ERROR();
                code = (0x1B<<16) | y<<8 | x;
            }
        }

        if(strcmp(aw.w0,"CALLNZ")==0) {
            if( n!=2 && n!=3 ) EXIT_SYNTAX_ERROR();
            if( n==2 ) {
                x = get_address(aw.w1);
                code = (0x1C<<16) | ((x&0x3f80)<<1) | (x&0x7f);
            } else {
                y = opr_7bit(asm_opr(aw.w1));
                x = opr_7bit(asm_opr(aw.w2));
                if(y&0x80) EXIT_SYNTAX_ERROR();
                code = (0x1C<<16) | y<<8 | x;
            }
        }

        if(strcmp(aw.w0,"CALLZ")==0) {
            if( n!=2 && n!=3 ) EXIT_SYNTAX_ERROR();
            if( n==2 ) {
                x = get_address(aw.w1);
                code = (0x1D<<16) | ((x&0x3f80)<<1) | (x&0x7f);
            } else {
                y = opr_7bit(asm_opr(aw.w1));
                x = opr_7bit(asm_opr(aw.w2));
                if(y&0x80) EXIT_SYNTAX_ERROR();
                code = (0x1D<<16) | y<<8 | x;
            }
        }

        if(strcmp(aw.w0,"RETURN")==0) {
            if( n!=1 ) EXIT_SYNTAX_ERROR();
            code = 0x20 << 16;
        }

        if(strcmp(aw.w0,"RETURNC0")==0) {
            if( n!=1 ) EXIT_SYNTAX_ERROR();
            code = 0x22 << 16;
        }

        if(strcmp(aw.w0,"RETURNC1")==0) {
            if( n!=1 ) EXIT_SYNTAX_ERROR();
            code = 0x23 << 16;
        }

        if(strcmp(aw.w0,"RETURNNZ")==0) {
            if( n!=1 ) EXIT_SYNTAX_ERROR();
            code = 0x24 << 16;
        }

        if(strcmp(aw.w0,"RETURNZ")==0) {
            if( n!=1 ) EXIT_SYNTAX_ERROR();
            code = 0x25 << 16;
        }

        if(strcmp(aw.w0,"LOAD&RETURN")==0) {
            if( n != 3 ) EXIT_SYNTAX_ERROR();

            z = opr_7bit(asm_opr(aw.w1));
            if( (z&0x80)==0 ) EXIT_SYNTAX_ERROR();

            x = asm_opr_load(aw.w2);
            if(x & 0x8000) {
                x = 0x80 | (0x7F&x);
                code = (z<<24) | (0x20<<16) | x;
            } else {
                if(x>255) exit_msg("LOAD constant range over");
                y = x/2;
                c = x&1;
                x = y;
                code = (z<<24) | (0x20<<16) | (y<<8) | x | (c<<16);
            }
        }

        if(strcmp(aw.w0,"INPUT")==0) {
            if( n != 3 ) EXIT_SYNTAX_ERROR();
            z = opr_7bit(asm_opr(aw.w1));
            if( (z&0x80)==0 ) EXIT_SYNTAX_ERROR();
            x = opr_7bit(asm_opr(aw.w2));
            code = (z<<24) | (0xC0<<16) | x;
        }

        if(strcmp(aw.w0,"INPUTAND")==0) {
            if( n != 4 ) EXIT_SYNTAX_ERROR();
            z = opr_7bit(asm_opr(aw.w1));
            if( (z&0x80)==0 ) EXIT_SYNTAX_ERROR();
            x = opr_7bit(asm_opr(aw.w2));
            y = opr_7bit(asm_opr(aw.w3));
            code = (z<<24) | (0xC1<<16) | (y<<8)| x;
        }

        if(strcmp(aw.w0,"INPUTXOR")==0) {
            if( n != 4 ) EXIT_SYNTAX_ERROR();
            z = opr_7bit(asm_opr(aw.w1));
            if( (z&0x80)==0 ) EXIT_SYNTAX_ERROR();
            x = opr_7bit(asm_opr(aw.w2));
            y = opr_7bit(asm_opr(aw.w3));
            code = (z<<24) | (0xC2<<16) | (y<<8)| x;
        }

        if(strcmp(aw.w0,"INPUTOR")==0) {
            if( n != 4 ) EXIT_SYNTAX_ERROR();
            z = opr_7bit(asm_opr(aw.w1));
            if( (z&0x80)==0 ) EXIT_SYNTAX_ERROR();
            x = opr_7bit(asm_opr(aw.w2));
            y = opr_7bit(asm_opr(aw.w3));
            code = (z<<24) | (0xC3<<16) | (y<<8)| x;
        }

        if(strcmp(aw.w0,"INPUTTEST")==0) {
            if( n != 3 ) EXIT_SYNTAX_ERROR();
            x = opr_7bit(asm_opr(aw.w1));
            y = opr_7bit(asm_opr(aw.w2));
            code = (0xC1<<16) | (y<<8)| x;
        }

        if(strcmp(aw.w0,"OUTPUT")==0) {
            if( n != 3 ) EXIT_SYNTAX_ERROR();
            x = opr_7bit(asm_opr(aw.w1));
            y = opr_7bit(asm_opr(aw.w2));
            code = (0x07<<16) | (y<<8) | x;
        }

        if(strcmp(aw.w0,"FETCH")==0) {
            if( n != 3 ) EXIT_SYNTAX_ERROR();
            x = opr_7bit(asm_opr(aw.w1));
            y = opr_7bit(asm_opr(aw.w2));
            code = (0x45<<16) | (y<<8) | x;
        }

        if(strcmp(aw.w0,"STORE")==0) {
            if( n != 4 ) EXIT_SYNTAX_ERROR();
            z = opr_7bit(asm_opr(aw.w1));
            if( (z&0x80)==0 ) EXIT_SYNTAX_ERROR();
            x = opr_7bit(asm_opr(aw.w2));
            y = opr_7bit(asm_opr(aw.w3));
            code = (z<<24) | (0x44<<16) | (y<<8) | x;
        }

        if(strcmp(aw.w0,"OUTPUT2")==0) {
            if( n != 3 ) EXIT_SYNTAX_ERROR();
            x = opr_7bit(asm_opr(aw.w1));
            y = opr_7bit(asm_opr(aw.w2));
            code = (0x46<<16) | (y<<8) | x;
        }

        if(strcmp(aw.w0,"EXIT")==0) {
            code = 0x00ff0000;
        }

        if(code==0 &&  buf[0] != '\n' && buf[1] != 'r') {
            fprintf(fp,"ERROR: invalid op in %d\n%s\n",err_line,buf);
            exit(-1);
        }

        if(code==0x80) code = 0;

        if(lwmode) code = ((code>>16)&0xff) | ((code<<8)&0x00ffff00) | (code&0xff000000);

        if(mode&M_STDOUT) {
            if(lwmode) {
                if(mode&M_DEBUG) fprintf(fp,"%04X %08X %016llX %s",adr_line,code,debug,buf);
                else fprintf(fp,"%04X %08X %s",adr_line,code,buf);
            } else {
                if(mode&M_DEBUG) fprintf(fp,"(%04X)%04X %08X %016llX %s",adr_line,adr_line*4,code,debug,buf);
                else fprintf(fp,"(%04X)%04X %08X %s",adr_line,adr_line*4,code,buf);
            }
        }
        PCODE[adr_line++] = code;
    }
    fclose(fasm);

    if(mode&M_BIN) {

        if(lwmode) {
            if(mode&M_BIN) {
                for(i=0 ; i<adr_line ; i++) fprintf(fp,"%08X\n", PCODE[i]);
                for( ; i<512 ; i++) fprintf(fp,"%08X\n",0);
            }
        } else {
            for(i=0 ; i<adr_line ; i++) {
                fprintf(fp,"%02X %02X %02X %02X\n",
                    PCODE[i]&0xff,(PCODE[i]>>8)&0xff,(PCODE[i]>>16)&0xff,(PCODE[i]>>24)&0xff);
            }
            for( ; i<512 ; i++) fprintf(fp,"%02X %02X %02X %02X\n",0,0,0,0);
        }
    }

}
